"""
android-sdk-updater

:copyright: (c) 2016 Tad Fisher
:license: Apache 2.0, see LICENSE for more details.

"""

__version__ = '1.0.1'
__author__ = 'Tad Fisher'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2016 Tad Fisher'



class Package:
    def __init__(self, category, name, revision, semver, num=0):
        self.category = category
        self.name = name
        self.revision = revision
        self.semver = semver
        self.num = num

    def __str__(self):
        return '{0.name} [{0.revision}]'.format(self)

    def __eq__(self, other):
        return self.name == other.name and self.semver == other.semver

    def __ne__(self, other):
        return not __eq__(other)

    def __hash__(self):
        const = 37
        total = 13
        total += self.name.__hash__() * total + const
        total += self.semver.__hash__() * total + const
        return total



#!/usr/bin/python

from __future__ import print_function

import os
import sys
import jprops

from semantic_version import Version

from sdk_updater.package import Package


# TODO: Support NDK installs
blacklist = [
    'ndk-bundle',
    'temp'
]


def add_ons(props, parts):
    return parts[1]


def build_tools(props, parts):
    return '-'.join(parts)


def docs(props, parts):
    return 'doc-' + props['api']


def extras(props, parts):
    return '-'.join(['extra'] + parts[1:])


def platforms(props, parts):
    return 'android-' + props['api']


def platform_tools(props, parts):
    return parts[0]


def samples(props, parts):
    return 'sample-' + props['api']


def sources(props, parts):
    return 'source-' + props['api']


def system_images(props, parts):
    tag = props['tag']
    if tag == 'default':
        tag = 'android'
    return '-'.join(['sys-img', props['abi'], tag, props['api']])


def tools(props, parts):
    return 'tools'


def default(props, parts):
    return None


def parse(top, root):
    path = os.path.relpath(root, top)

    parts = path.split(os.path.sep)

    if parts[0] in blacklist:
        print('WARNING: Ignoring \'{:s}\' as it is blacklisted.'.format(path), file=sys.stderr)
        return None

    props = parse_properties(os.path.join(root, 'source.properties'))
    name = {
        'add-ons': add_ons,
        'build-tools': build_tools,
        'docs': docs,
        'extras': extras,
        'platforms': platforms,
        'platform-tools': platform_tools,
        'samples': samples,
        'sources': sources,
        'system-images': system_images,
        'tools': tools
    }.get(parts[0], default)(props, parts)
    if not name:
        print("WARNING: Failed to parse package:", path, file=sys.stderr)
        return None
    return Package(parts[0], name, props['revision'], Version.coerce(props['revision']))


def parse_properties(file):
    props = {}
    with open(file, 'rb') as fp:
        for key, value in jprops.iter_properties(fp):
            if key == 'AndroidVersion.ApiLevel':
                props['api'] = value
            elif key == 'Pkg.LicenseRef':
                props['license'] = value
            elif key == 'Pkg.Revision':
                props['revision'] = value
            elif key == 'SystemImage.Abi':
                props['abi'] = value
            elif key == 'SystemImage.TagId':
                props['tag'] = value
    return props


def scan(top, verbose=False):
    packages = []
    for root, subdirs, files in os.walk(top, followlinks=True):
        if 'source.properties' in files:
            del subdirs[:]
            package = parse(top, root)
            if package:
                packages.append(package)
                if verbose:
                    print('Found package "{:s}" in {:s}'.format(str(package), root))
    return packages


def main(top):
    packages = scan(top)
    for package in packages:
        print(str(package))


if __name__ == '__main__':
    main(sys.argv[1])



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import sdk_updater

requires = [
    'jprops',
    'pexpect',
    'semantic_version',
]

setup(
    name='android-sdk-updater',
    version=sdk_updater.__version__,
    description='A command-line utility for keeping your Android dependencies up-to-date.',
    long_description=open('README.rst').read(),
    keywords='android',
    license=open('LICENSE').read(),
    author='Tad Fisher',
    author_email='tadfisher@gmail.com',
    url='https://github.com/tadfisher/android-sdk-updater',
    install_requires=requires,
    packages=['sdk_updater'],
    scripts=['bin/android-sdk-updater'],
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Build Tools'
    ]
)



#!/usr/bin/python

from __future__ import print_function
import os
import sys
import pexpect
from sdk_updater.scan import scan
from sdk_updater.list import list_packages


class PackageOperation(object):
    def __init__(self, package):
        self.package = package


class Install(PackageOperation):
    def __init__(self, package, revision):
        super(Install, self).__init__(package)
        self.revision = revision

    def __str__(self):
        return 'Installing: {:s} [{:s}]'\
            .format(self.package.name, self.revision)


class Update(PackageOperation):
    def __init__(self, package, revision_installed, revision_available):
        super(Update, self).__init__(package)
        self.revision_installed = revision_installed
        self.revision_available = revision_available

    def __str__(self):
        return 'Updating: {:s} [{:s} -> {:s}]'\
            .format(self.package.name, self.revision_installed, self.revision_available)


def to_dict(packages):
    d = {}
    for p in packages:
        d[p.name] = p
    return d


def compute_package_ops(package_names, installed, available):
    ops = []
    missing = set()

    ad = to_dict(available)
    installed_names = [p.name for p in installed]
    
    if package_names is not None:
        not_installed = set(package_names) - set(installed_names)

        for name in not_installed:
            p = ad.get(name)
            if p is None:
                missing.add(name)
            else:
                ops.append(Install(p, p.revision))

    for i in installed:
        p = ad.get(i.name)
        if p is None:
            missing.add(i.name)
        elif p.semver > i.semver:
            ops.append(Update(p, i.revision, p.revision))

    return ops, missing


def scan_missing(sdk, packages, verbose=False):
    # Re-scan SDK root to check for failed installs.
    print('Re-scanning {:s}...'.format(sdk))
    installed = scan(sdk, verbose=verbose)
    return [p for p in packages if p not in installed]


def install_packages(packages, android, options=None, verbose=False, timeout=None):
    if options is None:
        options = []
    package_filter = ','.join([p.num for p in packages])
    args = ['update', 'sdk', '--no-ui', '--all', '--filter', package_filter] + options
    installer = pexpect.spawn(android, args=args, timeout=timeout)
    if verbose:
        if sys.version_info >= (3,):
            installer.logfile = sys.stdout.buffer
        else:
            installer.logfile = sys.stdout
    while True:
        i = installer.expect([r"Do you accept the license '.+' \[y/n\]:",
                              pexpect.TIMEOUT, pexpect.EOF])
        if i == 0:
            # Prompt
            installer.sendline('y')
        elif i == 1:
            # Timeout
            print('WARNING: Package installation timed out after {:d} seconds.'
                  .format(timeout), file=sys.stderr)
            break
        else:
            break

def updates_available(android, sdk, packages=None, options=None, verbose=False):
    print('Scanning', sdk, 'for installed packages...')
    installed = scan(sdk, verbose=verbose)
    print('   ', str(len(installed)), 'packages installed.')

    print('Querying update sites for available packages...')
    available = list_packages(android, options=options, verbose=verbose)
    print('   ', str(len(available)), 'packages available.')

    return compute_package_ops(packages, installed, available)


def main(sdk, packages=None, options=None, verbose=False, timeout=None, dry_run=False):
    if timeout == 0:
        timeout = None

    android = os.path.join(sdk, 'tools', 'android')
    if not os.path.isfile(android):
        print('{:s} not found. Is ANDROID_HOME correct?'.format(android))
        exit(1)

    ops, missing_remote = updates_available(android, sdk, packages, options, verbose)

    if missing_remote:
        print('WARNING: Remote repository is missing package(s):', file=sys.stderr)
        for name in missing_remote:
            print('    ', name, file=sys.stderr)
        if packages:
            missing_requested = set(missing_remote).intersection(set(packages))
            if missing_requested:
                print('FATAL: Remote repository is missing explicitly requested package(s):', file=sys.stderr)
                for p in missing_requested:
                    print('    ', p, file=sys.stderr)
                exit(1)

    if not ops:
        print("All packages are up-to-date.")
        exit(0)

    if dry_run:
        print("--dry-run was set; exiting.")
        exit(0)

    for op in ops:
        print(op)

    to_install = [op.package for op in ops]

    install_packages(to_install, android, options, verbose, timeout)

    # Re-scan SDK dir for packages we missed.
    missing = scan_missing(sdk, to_install, verbose=verbose)

    if missing:
        print('FATAL: Finished: {:d} packages installed. Failed to install {:d} package(s):'
              .format(len(to_install) - len(missing), len(missing)), file=sys.stderr)
        for p in missing:
            print('    ', p, file=sys.stderr)
        exit(1)
    else:
        print('Finished: {:d} packages installed.'.format(len(to_install)))
        exit(0)

if __name__ == '__main__':
    main(sys.argv[1:])



#!/usr/bin/python

from __future__ import print_function
from __future__ import unicode_literals

import re
import sys
import subprocess

from semantic_version import Version

from sdk_updater.package import Package

categories = {
    'Addon': 'add-ons',
    'BuildTool': 'build-tools',
    'Doc': 'docs',
    'Extra': 'extras',
    'Platform': 'platforms',
    'PlatformTool': 'platform-tools',
    'Sample': 'samples',
    'Source': 'sources',
    'SystemImage': 'system-images',
    'Tool': 'tools'
}


def list_packages(android, options=None, verbose=False):
    if options is None:
        options = []
    packages = []
    separator = '----------'

    out = subprocess.check_output(
        [android, 'list', 'sdk', '--all', '--extended'] + options,
        stderr=subprocess.PIPE)

    out = out.decode('utf-8')

    if verbose:
        print(out)

    fields = out.split(separator)[1:]
    p_id = re.compile('^id: (\d+) or "(.+)"$', flags=re.MULTILINE)
    p_revision = re.compile('[Rr]evision (.+)')
    p_type = re.compile('Type: (\w+)')
    for field in fields:
        m = p_id.search(field)
        if m is None:
            print("WARNING: Failed to parse package ID:", field, file=sys.stderr)
            continue
        num, name = m.groups()
        m = p_revision.search(field)
        if m is None:
            print("WARNING: Failed to parse revision:", field, file=sys.stderr)
            continue
        revision, = m.groups()
        revision = revision.replace(' (Obsolete)', '')
        semver = Version.coerce(revision)

        m = p_type.search(field)
        if m is None:
            print('WARNING: Failed to parse type:', field, file=sys.stderr)
            continue
        ptype, = m.groups()
        category = categories[ptype]
        if category is None:
            print('WARNING: Unrecognized type:', ptype, file=sys.stderr)
            category = ptype.lower()
        packages.append(Package(category, name, revision, semver, num))
    return packages


def main(android):
    packages = list_packages(android)
    for p in packages:
        print(p)

if __name__ == '__main__':
    main(sys.argv[1])
